"""
Copyright 2020 Skyscanner Ltd

Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import json
from lambdaguard.utils.arnparse import arnparse
from lambdaguard.utils.log import debug
from lambdaguard.core.AWS import AWS
from lambdaguard.core.Role import Role
from lambdaguard.core.KMS import KMS
from lambdaguard.security.Scan import Scan


class Lambda(AWS):
    def __init__(self, arn, *args, **kwargs):
        super().__init__(arn, args[0].profile, args[0].keys[0], args[0].keys[1])

        self.args = args[0]
        self.identity = args[1]
        self.runtime = None
        self.handler = None
        self.layers = None
        self.description = None
        self.role = None
        self.kms = None
        self.codeURL = None
        self.writes = {'count': 0, 'items': {}}
        self.triggers = {'services': [], 'items': {}}
        self.resources = {'services': [], 'items': {}}
        self.security = {'count': {}, 'items': {}}

        self.get_function()
        self.get_policy()
        self.get_triggers()
        self.get_resources()
        self.get_security()

    def get_policy(self):
        '''
        Fetches Function (Resource-based) policy
        '''
        try:
            policy = self.client.get_policy(FunctionName=self.arn.resource)
            self.policy = json.loads(policy['Policy'])
        except Exception:
            debug(self.arn.full)

    def get_function(self):
        '''
        Fetches Lambda function configuration
        '''
        try:
            if self.identity.acl.allowed("lambda:GetFunction"):
                function = self.client.get_function(
                    FunctionName=self.arn.resource
                )
                config = function['Configuration']
                self.codeURL = function['Code']['Location']
            elif self.identity.acl.allowed("lambda:GetFunctionConfiguration"):
                config = self.client.get_function_configuration(
                    FunctionName=self.arn.resource
                )
                self.codeURL = ''
            else:
                exit("\nMissing both lambda:GetFunction and lambda:GetFunctionConfiguration")

            self.runtime = config['Runtime']
            self.handler = config['Handler']
            self.description = config['Description']
            if 'KMSKeyArn' in config:
                self.kms = KMS(
                    config['KMSKeyArn'],
                    profile=self.profile,
                    access_key_id=self.access_key_id,
                    secret_access_key=self.secret_access_key
                )
            self.role = Role(
                config['Role'],
                profile=self.profile,
                access_key_id=self.access_key_id,
                secret_access_key=self.secret_access_key
            )
            self.layers = []
            if 'Layers' in config:
                for layer in config['Layers']:
                    layer = self.client.get_layer_version_by_arn(Arn=layer['Arn'])
                    self.layers.append({
                        'arn': layer['LayerVersionArn'],
                        'description': layer['Description'],
                        'codeURL': layer['Content']['Location']
                    })
        except Exception:
            debug(self.arn.full)

    def get_triggers(self):
        '''
        Tracks events that trigger the Lambda function
        '''
        # Collect triggers from Event Sources
        try:
            eventSource = self.client.list_event_source_mappings(
                FunctionName=self.arn.resource
            )['EventSourceMappings']

            for event in eventSource:
                if event['State'] != 'Enabled':
                    continue

                self.triggers['items'][event['EventSourceArn']] = ['lambda:InvokeFunction']

                # Track services
                self.triggers['services'].append(arnparse(event['EventSourceArn']).service)
        except Exception:
            debug(self.arn.full)

        # Collect triggers from Function policy
        try:
            if self.policy:
                for statement in self.policy['Statement']:
                    if 'Condition' in statement:
                        if 'ArnLike' in statement['Condition']:
                            if 'AWS:SourceArn' in statement['Condition']['ArnLike']:
                                arn = statement['Condition']['ArnLike']['AWS:SourceArn']

                                if type(statement['Action']) == str:
                                    self.triggers['items'][arn] = [statement['Action']]
                                else:
                                    self.triggers['items'][arn] = statement['Action']

                                # Track services
                                self.triggers['services'].append(arnparse(arn).service)
        except Exception:
            debug(self.arn.full)

        self.triggers['services'] = list(set(self.triggers['services']))

    def get_resources(self):
        '''
        Tracks resources used by the Lambda function
        '''
        try:
            if self.role:
                for policy in self.role.policy['policies']:
                    for statement in policy['document']['Statement']:
                        if type(statement) != dict:
                            continue
                        if statement['Effect'] != 'Allow':
                            continue

                        if type(statement['Resource']) == str:
                            arns = [statement['Resource']]
                        else:
                            arns = statement['Resource']

                        if type(statement['Action']) == str:
                            actions = [statement['Action']]
                        else:
                            actions = statement['Action']

                        # Track services
                        for action in actions:
                            self.resources['services'].append(
                                action.split(':')[0]
                            )

                        # Track actions by resource
                        for arn in arns:
                            if arn in self.resources['items']:
                                self.resources['items'][arn] = list(
                                    set(self.resources['items'][arn] + actions)
                                )
                            else:
                                self.resources['items'][arn] = actions

            self.resources['services'] = list(set(self.resources['services']))
        except Exception:
            debug(self.arn.full)

    def get_security(self):
        try:
            self.security = Scan(
                self.report(),
                self.args
            ).security
        except Exception:
            debug(self.arn.full)

    def set_writes(self, writes):
        try:
            self.writes['count'] += 1
            for arn, policy in writes.items():
                self.writes['items'][arn] = policy
        except Exception:
            debug(self.arn.full)

    def report(self):
        ret = {
            'arn': self.arn.full,
            'name': self.arn.resource,
            'description': self.description,
            'region': self.arn.region,
            'runtime': self.runtime,
            'handler': self.handler,
            'layers': self.layers,
            'codeURL': self.codeURL,
            'role': self.role.arn.full,
            'policy': {
                'function': self.policy,
                'role': self.role.policy
            },
            'writes': self.writes,
            'triggers': self.triggers,
            'resources': self.resources,
            'security': self.security
        }
        if self.kms:
            ret['kms'] = self.kms.arn.full
            ret['policy']['kms'] = self.kms.policies
        return ret



"""
Copyright 2020 Skyscanner Ltd

Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


class AccessControlList:
    def __init__(self, acl):
        self.acl = acl

    def audit(self):
        uris = {
            'http://acs.amazonaws.com/groups/global/AllUsers': 'Everyone',
            'http://acs.amazonaws.com/groups/global/AuthenticatedUsers': 'Authenticated AWS users'
        }

        if not self.acl:
            yield {
                'level': 'info',
                'text': 'Bucket ACL is not defined'
            }
        else:
            for acl in self.acl['Grants']:
                if acl['Grantee']['Type'] == 'Group':
                    if acl['Grantee']['URI'] in uris:
                        yield {
                            'level': 'high',
                            'text': f'Public Bucket ACL: {acl["Permission"]} access for {uris[acl["Grantee"]["URI"]]}'
                        }



"""
Copyright 2020 Skyscanner Ltd

Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import boto3
from pathlib import Path
from shutil import rmtree
from lambdaguard.utils.arnparse import arnparse
from lambdaguard.utils.log import debug
from lambdaguard.utils.cli import parse_args, align, header, nocolor, green, orange
from lambdaguard.utils.log import configure_log
from lambdaguard.utils.paginator import paginate
from lambdaguard.core.Lambda import Lambda
from lambdaguard.core.STS import STS
from lambdaguard.visibility.Statistics import Statistics
from lambdaguard.visibility.Report import VisibilityReport
from lambdaguard.visibility.HTMLReport import HTMLReport
from lambdaguard.security.LambdaWrite import LambdaWrite
from lambdaguard.security.Report import SecurityReport


def verbose(args, message, end=''):
    '''
    Prints formatted message if verbose mode is set
    '''
    if args.verbose:
        print(
            f'\r{green}{message}{nocolor}'.ljust(100, ' '),
            end=end
        )


def get_client(args):
    '''
    Returns a Lambda botocore client
    '''
    return boto3.Session(
        profile_name=args.profile,
        aws_access_key_id=args.keys[0],
        aws_secret_access_key=args.keys[1],
        region_name=args.region
    ).client('lambda')


def get_regions(args):
    '''
    Valid region specification:
        Single:     eu-west-1
        Multiple:   eu-west-1,ap-south-1,us-east-2
        All:        all
    Returns regions as a Python list
    '''
    if not isinstance(args.region, str):
        raise ValueError(f'No region specified')
    if args.function:
        return [arnparse(args.function).region]
    available = boto3.Session().get_available_regions('lambda')
    if args.region == 'all':
        return available
    regions = args.region.split(',')
    if not regions:
        raise ValueError(f'No region specified')
    for region in regions:
        if region not in available:
            raise ValueError(f'Invalid region "{region}"')
    return regions


def get_usage(args):
    '''
    Returns Python dict with number of Lambdas per region
    '''
    usage = {}
    for region in get_regions(args):
        args.region = region
        verbose(args, f'Loading regions ({region})')
        try:
            client = get_client(args)
            settings = client.get_account_settings()
            function_count = settings['AccountUsage']['FunctionCount']
            if function_count:
                usage[region] = function_count
        except Exception:
            debug(region)
    return usage


def get_functions(args):
    '''
    Generator for listing Lambda functions
    Yields Lambda function ARNs
    '''
    if args.function:
        yield args.function
    elif args.input:
        with Path(args.input).open() as f:
            for _ in f.read().split('\n'):
                yield _
    else:
        client = get_client(args)
        for page in paginate(client, 'list_functions'):
            for function in page['Functions']:
                yield function['FunctionArn']


def run(arguments=''):
    '''
    Main routine
    '''
    args = parse_args(arguments)

    verbose(args, header, end='\n\n')

    if args.html:
        HTMLReport(args.output).save()
        verbose(args, f'Generated {args.output}/report.html', end='\n\n')
        exit(0)

    rmtree(args.output, ignore_errors=True)
    Path(args.output).mkdir(parents=True, exist_ok=True)
    configure_log(args.output)
    usage = get_usage(args)
    verbose(args, f'Loading identity')
    region = list(usage.keys())[0]
    sts_arn = f'arn:aws:sts:{region}'
    identity = STS(sts_arn, args.profile, args.keys[0], args.keys[1])
    if args.verbose:
        for _ in ['UserId', 'Account', 'Arn']:
            align(_, identity.caller[_], orange)
        print('')

    statistics = Statistics(args.output)
    visibility = VisibilityReport(args.output)
    writes = LambdaWrite(args)
    total_count = 0
    for region_count in usage.values():
        total_count += region_count

    for region in usage.keys():
        args.region = region
        for arn_str in get_functions(args):
            try:
                arn = arnparse(arn_str)
                counter = f'[ {statistics.statistics["lambdas"]+1}/{total_count} ] '
                verbose(args, f'{counter}{arn.resource}')
                lmbd = Lambda(arn.full, args, identity)
                for w in writes.get_for_lambda(arn.full):
                    lmbd.set_writes(w)
                statistics.parse(lmbd.report())
                visibility.save(lmbd.report())
            except Exception:
                debug(arn_str)

    SecurityReport(args.output).save()
    HTMLReport(args.output).save()

    if args.verbose:
        print('\r' + ' ' * 100, end='\r')  # clear
        align('Lambdas', statistics.statistics["lambdas"])
        align('Security', statistics.statistics["security"]["count"])
        align('Triggers', statistics.statistics["triggers"]["count"])
        align('Resources', statistics.statistics["resources"]["count"])
        align('Layers', statistics.statistics["layers"])
        align('Runtimes', len(statistics.statistics["runtimes"]["items"]))
        align('Regions', len(statistics.statistics["regions"]["items"]))
        print('')
        align('Report', f'{args.output}/report.html')
        align('Log', f'{args.output}/lambdaguard.log')
        print('')



"""
Copyright 2020 Skyscanner Ltd

Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import boto3
from lambdaguard.utils.arnparse import arnparse


class AWS(object):
    '''
    Base AWS service object extended by each individual service class.
    '''
    def __init__(self, arn, profile=None, access_key_id=None, secret_access_key=None):
        # AWS ARN
        self.arn = arnparse(arn)

        # AWS Profile and Keys
        self.profile = profile
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key

        # AWS Resource-based policy
        self.policy = {}

        # Additional service information
        self.info = ''

        # AWS connection
        session = boto3.Session(profile_name=self.profile)
        self.client = session.client(
            self.arn.service,
            region_name=self.arn.region,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key
        )



"""
Copyright 2020 Skyscanner Ltd

Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import logging
import traceback
import logging.config
from pathlib import Path


def configure_log(path=''):
    logpath = Path(path, 'lambdaguard.log').as_posix()
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'default': {
                'class': 'logging.Formatter',
                'style': '{',
                'datefmt': '%Y-%m-%d %H:%M',
                'format': '[{asctime:s}] {message:s}\n'
            }
        },
        'handlers': {
            'lambdaguard-log': {
                'level': 'DEBUG',
                'class': 'logging.handlers.WatchedFileHandler',
                'formatter': 'default',
                'filename': logpath,
                'mode': 'w',
                'encoding': 'utf-8'
            }
        },
        'loggers': {},
        'root': {
            'handlers': ['lambdaguard-log'],
            'level': 'WARNING'
        }
    })


def debug(arn=''):
    # Get exception name and description
    trace = traceback.format_exc().strip()
    etype, evalue = trace.split('\n')[-1].split(': ', 1)

    # Drop known messages
    if 'botocore.e' in trace:
        # Differently formatted exception messages
        if not etype.startswith('botocore'):
            etype, evalue = trace.split('\n')[-2].split(': ', 1)

        if etype == 'botocore.errorfactory.NoSuchEntityException':
            return None  # Missing policy
        elif etype == 'botocore.errorfactory.ResourceNotFoundException':
            return None  # Missing resource
        elif etype == 'botocore.exceptions.ParamValidationError':
            return None  # Invalid resource name
        elif etype == 'botocore.errorfactory.NoSuchBucket':
            return None  # Missing S3 bucket
        elif etype == 'botocore.errorfactory.NotFoundException':
            return None  # Missing SNS topic
        elif etype == 'botocore.errorfactory.QueueDoesNotExist':
            return None  # Missing SQS queue
        elif etype == 'botocore.exceptions.ClientError':
            if 'AccessDenied' in evalue:
                # TODO: track denied resources
                return None
            elif 'NoSuchBucketPolicy' in evalue:
                return None  # Missing S3 bucket policy
            elif 'ServerSideEncryptionConfigurationNotFoundError' in evalue:
                return None  # Missing S3 bucket encryption
            elif 'ValidationException' in evalue:
                return None  # Invalid resource value
            elif 'arn:aws:lambda:::awslayer:' in evalue:
                return None  # Update opt-in/out
    elif trace.startswith('NoneType: None'):
        return None  # Empty
    elif etype == 'ValueError':
        if evalue.startswith('Invalid endpoint:'):
            return None  # Invalid resource configuration

    # Log
    logging.warning(f'[{arn}]\n{trace}')
    return trace



"""
Copyright 2020 Skyscanner Ltd

Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


class PrivilegeEscalation:
    def __init__(self, item):
        self.item = item

    def audit(self):
        exploitable = []
        iam = [
            'iam:*',
            'iam:AddUserToGroup',
            'iam:AttachGroupPolicy',
            'iam:AttachRolePolicy',
            'iam:AttachUserPolicy',
            'iam:CodeStarCreateProjectFromTemplate',
            'iam:CodeStarCreateProjectThenAssociateTeamMember',
            'iam:CreateAccessKey',
            'iam:CreateEC2WithExistingIP',
            'iam:CreateLoginProfile',
            'iam:CreateNewPolicyVersion',
            'iam:EditExistingLambdaFunctionWithRole',
            'iam:PassExistingRoleToNewCloudFormation',
            'iam:PassExistingRoleToNewCodeStarProject',
            'iam:PassExistingRoleToNewDataPipeline',
            'iam:PassExistingRoleToNewGlueDevEndpoint',
            'iam:PassExistingRoleToNewLambdaThenInvoke',
            'iam:PassExistingRoleToNewLambdaThenTriggerWithExistingDynamo',
            'iam:PassExistingRoleToNewLambdaThenTriggerWithNewDynamo',
            'iam:PutGroupPolicy',
            'iam:PutRolePolicy',
            'iam:PutUserPolicy',
            'iam:SetExistingDefaultPolicyVersion',
            'iam:UpdateExistingGlueDevEndpoint',
            'iam:UpdateLoginProfile',
            'iam:UpdateRolePolicyToAssumeIt'
        ]

        for action in self.item:
            if action in iam:
                exploitable.append(action)

        if len(exploitable):
            exploitable = ', '.join(exploitable)
            yield {
                'level': 'high',
                'text': f'Potential privilege escalation via the following Actions: {exploitable}'
            }



"""
Copyright 2020 Skyscanner Ltd

Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import json
from base64 import b64encode
from pathlib import Path


class HTMLReport:
    def __init__(self, path):
        self.path = Path(path)

    def txt2html(self, text):
        html = text.split('\n')
        return '<br>'.join(html)

    def save(self):
        '''
        Generates an HTML report
        '''
        assets_path = Path(__file__).parents[1].joinpath('assets')

        with assets_path.joinpath('index.html').open() as f:
            report_html = f.read()

        with assets_path.joinpath('logo.png').open('rb') as f:
            logo = b64encode(f.read()).decode('ascii')
            report_html = report_html.replace('{logo}', str(logo))

        # ===================================================== Statistics
        if not self.path.joinpath('statistics.json').exists():
            return
        with self.path.joinpath('statistics.json').open() as f:
            stats_json = json.loads(f.read())

        exclude_info = stats_json['lambdas'] > 1  # Exclude info level

        stats_table = assets_path.joinpath('stats.html').open().read()
        stats_table_item = assets_path.joinpath('statsitem.html').open().read()
        for title, data in stats_json.items():
            if title in ['lambdas', 'regions', 'layers']:
                continue

            if exclude_info and title == 'security':
                count = data['count'] - data['items']['info']
            else:
                count = data['count']
            stats_table = stats_table.replace('{' + title + '_count}', str(count))

            stats_items = []
            for name, count in data['items'].items():
                if exclude_info and name == 'info':
                    continue
                stats_items.append(
                    stats_table_item.replace('{name}', name).replace('{count}', str(count))
                )
            stats_table = stats_table.replace('{' + title + '_items}', ''.join(stats_items))
        stats_table = stats_table.replace('{lambdas_count}', str(stats_json['lambdas']))
        stats_table = stats_table.replace('{layers_count}', str(stats_json['layers']))

        with assets_path.joinpath('highcharts.js').open() as f:
            report_html = report_html.replace('{highcharts}', f.read())
        with assets_path.joinpath('highcharts-theme.js').open() as f:
            report_html = report_html.replace('{highcharts-theme}', f.read())

        with assets_path.joinpath('stats.js').open() as f:
            stats_template = f.read()

        stats_js = []
        for title, data in stats_json.items():
            if title in ['lambdas', 'regions', 'layers']:
                continue

            series = []
            for name, y in data['items'].items():
                if exclude_info and name == 'info':
                    data['count'] -= y
                    continue
                series.append({'name': name, 'y': y})

            chart = stats_template
            chart = chart.replace('{chartid}', f'stats-{title}')
            chart = chart.replace('{title}', f'{title} ({data["count"]})')
            chart = chart.replace('{count}', str(data['count']))
            chart = chart.replace('{data}', json.dumps(series))
            stats_js.append(chart)

        report_html = report_html.replace('{statistics}', stats_table)
        report_html = report_html.replace('{highcharts-data}', '\n'.join(stats_js))

        stats_table = None
        stats_js = None

        # ===================================================== Security
        with self.path.joinpath('security.json').open() as f:
            vulns_json = json.loads(f.read())

        with assets_path.joinpath('vuln.html').open() as f:
            vuln_html_template = f.read()

        vulns_html = ''
        for _ in vulns_json:
            if exclude_info and _['level'] == 'info':
                continue
            html = vuln_html_template
            html = html.replace('{index}', _['index'])
            html = html.replace('{level}', _['level'])
            html = html.replace('{text}', self.txt2html(_['text']))
            html = html.replace('{where}', self.txt2html(_['where']))
            html = html.replace('{lambda}', _['lambda'])
            vulns_html += f'{html}\n'

        with assets_path.joinpath('vulnlist.html').open() as f:
            vulnlist_html = f.read()

        vulnlist_html = vulnlist_html.replace('{items}', vulns_html)

        report_html = report_html.replace('{security}', vulnlist_html)

        vulns_json = None
        vulns_html = None
        vulnlist_html = None

        # ===================================================== Functions & Layers
        all_layers = []

        with self.path.joinpath('index.json').open() as f:
            index = json.loads(f.read())

        with assets_path.joinpath('func.html').open() as f:
            func_html_template = f.read()

        with assets_path.joinpath('funcvuln.html').open() as f:
            vuln_html_template = f.read()

        funcs_html = ''

        for idx in index.keys():
            with self.path.joinpath('reports', f'{idx}.json').open() as f:
                func = json.loads(f.read())

            funcvuln_html = []

            for _ in func['security']['items']:
                if exclude_info and _['level'] == 'info':
                    continue
                ret = vuln_html_template.replace('{where}', self.txt2html(_['where']))
                ret = ret.replace('{level}', _['level'])
                ret = ret.replace('{text}', self.txt2html(_['text']))
                funcvuln_html.append(ret)

            layers = []
            for layer in func['layers']:
                layers.append(f"{layer['arn']} ({layer['description']})")
                layer.update({'runtime': func['runtime']})
                if layer not in all_layers:
                    all_layers.append(layer)

            html = func_html_template
            html = html.replace('{index}', idx)
            html = html.replace('{arn}', func['arn'])
            html = html.replace('{name}', func['name'])
            html = html.replace('{description}', func['description'])
            html = html.replace('{region}', func['region'])
            html = html.replace('{runtime}', func['runtime'])
            html = html.replace('{handler}', func['handler'])
            html = html.replace('{role}', func['role'])
            html = html.replace('{layers}', self.txt2html('\n'.join(layers)))
            html = html.replace('{layers_count}', str(len(func['layers'])))
            html = html.replace('{triggers}', self.txt2html(', '.join(func['triggers']['services'])))
            html = html.replace('{triggers_count}', str(len(func['triggers']['services'])))
            html = html.replace('{resources}', self.txt2html(', '.join(func['resources']['services'])))
            html = html.replace('{resources_count}', str(len(func['resources']['services'])))
            html = html.replace('{security}', ''.join(funcvuln_html))
            html = html.replace('{security_count}', str(len(funcvuln_html)))
            html = html.replace('{policy}', json.dumps(func['policy']['function'], indent=4))
            html = html.replace('{role_policy}', json.dumps(func['policy']['role'], indent=4))
            if 'kms' in func:
                html = html.replace('{kms}', func['kms'])
                html = html.replace('{kms_policies}', json.dumps(func['policy']['kms'], indent=4))
            else:
                html = html.replace('{kms}', '')
                html = html.replace('{kms_policies}', '{}')

            funcs_html += f'{html}\n'

        with assets_path.joinpath('funclist.html').open() as f:
            funclist_html = f.read()

        funclist_html = funclist_html.replace('{items}', funcs_html)

        report_html = report_html.replace('{functions}', funclist_html)

        funcs_html = None
        funclist_html = None

        with assets_path.joinpath('layer.html').open() as f:
            layer_html_template = f.read()

        layer_html = ''

        for layer in all_layers:
            html = layer_html_template.replace('{arn}', layer['arn'])
            html = html.replace('{description}', layer['description'])
            html = html.replace('{runtime}', layer['runtime'])
            layer_html += f'{html}\n'

        with assets_path.joinpath('layerlist.html').open() as f:
            layerlist_html = f.read()

        layerlist_html = layerlist_html.replace('{items}', layer_html)

        report_html = report_html.replace('{layers}', layerlist_html)

        layer_html = None
        layerlist_html = None

        # ===================================================== HTML

        with self.path.joinpath('report.html').open('w') as f:
            f.write(report_html)



"""
Copyright 2020 Skyscanner Ltd

Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import boto3
from lambdaguard.utils.iterator import iterate
from lambdaguard.utils.paginator import paginate


LAMBDA_WRITE_PERMISSIONS = [x.lower() for x in [
    '*',
    'lambda:*',
    'lambda:Create*',
    'lambda:Delete*',
    'lambda:Invoke*',
    'lambda:Publish*',
    'lambda:Put*',
    'lambda:Tag*',
    'lambda:Untag*',
    'lambda:Update*',
    'lambda:CreateAlias',
    'lambda:CreateFunction',
    'lambda:DeleteAlias',
    'lambda:DeleteEventSourceMapping',
    'lambda:DeleteFunction',
    'lambda:DeleteFunctionConcurrency',
    'lambda:DeleteLayerVersion',
    'lambda:InvokeAsync',
    'lambda:InvokeFunction',
    'lambda:PublishLayerVersion',
    'lambda:PublishVersion',
    'lambda:PutFunctionConcurrency',
    'lambda:TagResource',
    'lambda:UntagResource',
    'lambda:UpdateAlias',
    'lambda:UpdateEventSourceMapping',
    'lambda:UpdateFunctionCode',
    'lambda:UpdateFunctionConfiguration'
]]


def is_write_action(action):
    return action.lower() in LAMBDA_WRITE_PERMISSIONS


class LambdaWrite:
    """
    This is a class for tracking attached IAM policies
    that have Lambda Write permissions
    """
    def __init__(self, args):
        self.args = args
        self.writes = {}
        self.get_writes()

    def get_writes(self):
        for policy_arn, policy in self.get_attached_local_policies():
            for lambda_arn, actions in self.parse(policy):
                if lambda_arn not in self.writes:
                    self.writes[lambda_arn] = {}
                if policy_arn not in self.writes[lambda_arn]:
                    self.writes[lambda_arn][policy_arn] = {}
                self.writes[lambda_arn][policy_arn] = actions

    def get_attached_local_policies(self):
        client = boto3.Session(
            profile_name=self.args.profile,
            aws_access_key_id=self.args.keys[0],
            aws_secret_access_key=self.args.keys[1],
            region_name=self.args.region
        ).client('iam')
        pages = paginate(
            client,
            'list_policies',
            Scope='Local',
            OnlyAttached=True
        )
        for page in pages:
            for policy in page['Policies']:
                version = client.get_policy_version(
                    PolicyArn=policy['Arn'],
                    VersionId=policy['DefaultVersionId']
                )['PolicyVersion']
                yield policy['Arn'], version

    def parse(self, policy):
        if 'Document' not in policy:
            return StopIteration
        if 'Statement' not in policy['Document']:
            return StopIteration

        if not isinstance(policy['Document']['Statement'], list):
            statements = [policy['Document']['Statement']]
        else:
            statements = policy['Document']['Statement']

        for statement in statements:
            if 'Effect' not in statement:
                continue  # Unknown
            if statement['Effect'] != 'Allow':
                continue  # Not allowed
            if 'Action' not in statement:
                continue  # No permissions

            # Identify all write Actions
            write_actions = []
            for action in iterate(statement['Action']):
                if not is_write_action(action):
                    continue
                write_actions.append(action)
            if not write_actions:
                return StopIteration

            # Return all write Actions per Resource
            for arn in iterate(statement['Resource']):
                yield arn, write_actions

    def get_for_lambda(self, arn):
        for w_arn, w_actions in self.writes.items():
            if w_arn == '*':
                yield w_actions
            elif w_arn == arn:
                yield w_actions



"""
Copyright 2020 Skyscanner Ltd

Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import json
from lambdaguard.utils.log import debug
from lambdaguard.core.AWS import AWS


class S3(AWS):
    def __init__(self, arn, profile=None, access_key_id=None, secret_access_key=None):
        super().__init__(arn, profile, access_key_id, secret_access_key)

        self.acl = None
        self.encryption = None

        self.get_bucket_policy()
        self.get_bucket_acl()
        self.get_bucket_encryption()

        self.info = (
            f'http://{self.arn.resource}.s3.amazonaws.com\n\n' +
            f'https://console.aws.amazon.com/s3/buckets/{self.arn.resource}/?tab=permissions'
        )

    def get_bucket_policy(self):
        '''
        Fetches S3 Bucket (Resource-based) policy
        '''
        try:
            policy = self.client.get_bucket_policy(Bucket=self.arn.resource)
            self.policy = json.loads(policy['Policy'])
        except Exception:
            debug(self.arn.full)

    def get_bucket_acl(self):
        '''
        Fetches S3 Bucket (Resource-based) Access Control List
        '''
        try:
            self.acl = self.client.get_bucket_acl(Bucket=self.arn.resource)
        except Exception:
            debug(self.arn.full)

    def get_bucket_encryption(self):
        '''
        Fetches S3 Bucket encryption settings
        '''
        try:
            self.encryption = self.client.get_bucket_encryption(Bucket=self.arn.resource)
        except Exception:
            debug(self.arn.full)



"""
Copyright 2020 Skyscanner Ltd

Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
from setuptools import setup, find_packages
from pathlib import Path


def get_version():
    ret = {}
    code = Path(__file__).parent.joinpath('lambdaguard', '__version__.py')
    exec(code.read_text(), ret)
    return ret['__version__']


long_description = '''
LambdaGuard is an AWS Serverless Security auditing tool designed to create asset visibility 
and provide actionable results. It provides a meaningful overview in terms of 
statistical analysis, AWS service dependencies and configuration checks from 
the security perspective.

https://github.com/Skyscanner/LambdaGuard
'''

install_requires=[
    'boto3',
    'argparse',
    'requests'
]

dev_requires = [
    'coverage~=4.4',
    'flake8>=2.5.4',
    'pytest>=2.9.1',
    'pytest-mock>=1.0',
    'pip-tools',
    'wheel',
    'twine'
]

setup(
    name='LambdaGuard',
    version=get_version(),
    author='Artëm Tsvetkov',
    author_email='artem.tsvetkov@skyscanner.net',
    description='LambdaGuard',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Skyscanner/LambdaGuard',
    packages=find_packages(),
    include_package_data=True,
    # classifiers=[
    #     'Programming Language :: Python :: 3.6.3'
    # ],
    python_requires='>=3.6.3',
    setup_requires=['pytest-runner'],
    install_requires=install_requires,
    tests_require=dev_requires,
    extras_require={
        'dev': dev_requires,
    },
    entry_points={
        'console_scripts': ['lambdaguard=lambdaguard:run'],
    }
)



"""
Copyright 2020 Skyscanner Ltd

Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import json
import requests
from pathlib import Path
from shutil import rmtree
from zipfile import ZipFile, is_zipfile
from subprocess import check_call as sh
from subprocess import DEVNULL
from shlex import split as shsplit
from os import chdir as cd
from time import sleep


SONAR_PROJECT_PROPERTIES = '''
sonar.host.url={}
sonar.login={}
sonar.password={}
sonar.language={}
sonar.projectKey=lambdaguard
sonar.projectName=lambdaguard
sonar.sourceEncoding=UTF-8
sonar.sources=.
sonar.java.binaries=.
sonar.tests=
'''


class SonarQube:
    def __init__(self, config, output):
        self.downloads = Path(output, 'downloads')
        self.downloads.mkdir(parents=True, exist_ok=True)
        config = Path(config)
        if config.exists():
            self.config = json.loads(config.read_text())
        else:
            raise FileNotFoundError('SonarQube config file not found: {}'.format(config))

    def __del__(self):
        rmtree(self.downloads, ignore_errors=True)

    def scan(self, codeURL, runtime):
        if not self.config:
            return  # invalid config
        zippath = self.downloads.joinpath('lambda.zip')
        zippath.write_bytes(requests.get(codeURL).content)
        if not is_zipfile(zippath):
            return  # invalid zip
        zf = ZipFile(zippath)

        # Unzip Lambda source code
        for _ in zf.namelist():
            zf.extractall(self.downloads, members=[_])

        # Configure sonar-project.properties
        if runtime.startswith('python'):
            language = 'py'
        elif runtime.startswith('node'):
            language = 'js'
        elif runtime.startswith('java'):
            language = 'java'
        else:
            return  # unsupported language
        Path(self.downloads, 'sonar-project.properties').write_text(
            SONAR_PROJECT_PROPERTIES.format(
                self.config['url'],
                self.config['login'],
                self.config['password'],
                language
            )
        )

        # Run sonar-scanner
        cwd = Path('.').resolve()
        cd(self.downloads)
        sh(shsplit(self.config['command']), stdout=DEVNULL, stderr=DEVNULL)
        cd(cwd)
        rmtree(self.downloads, ignore_errors=True)
        self.downloads.mkdir(parents=True, exist_ok=True)

        # Get results
        curl = requests.Session()
        curl.auth = (self.config['login'], self.config['password'])

        while True:
            sleep(3)
            task = json.loads(
                curl.get(f'{self.config["url"]}/api/ce/activity').text
            )['tasks'][0]
            if task['status'] in ['SUCCESS', 'FAIL']:
                break

        issues = json.loads(
            curl.get(f'{self.config["url"]}/api/issues/search?project=lambdaguard').text
        )['issues']
        curl.post(f'{self.config["url"]}/api/projects/delete', data={
            'project': 'lambdaguard'
        })

        for issue in issues:
            if issue['status'] != 'OPEN':
                continue
            where = issue['component'].split(':', 1)[1]
            yield {
                'level': 'high',
                'text': f'{issue["message"]}\n{where} on line {issue["textRange"]["startLine"]}.'
            }
